package com.jqtx.windows.common.config;

import com.jqtx.windows.common.exception.BizException;
import com.jqtx.windows.infrastructure.enums.ExceptionCodeEnum;
import com.jqtx.windows.web.response.JsonResult;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.propertyeditors.CustomDateEditor;
import org.springframework.validation.BindException;
import org.springframework.validation.FieldError;
import org.springframework.web.HttpRequestMethodNotSupportedException;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.ServletRequestBindingException;
import org.springframework.web.bind.WebDataBinder;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.method.annotation.MethodArgumentTypeMismatchException;
import javax.servlet.http.HttpServletRequest;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;

@ControllerAdvice
@Slf4j
public class ExceptionConfig {

    @InitBinder
    public void initBinder(WebDataBinder binder) {
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
        binder.registerCustomEditor(Date.class, new CustomDateEditor(dateFormat, true));
        // true:允许输入空值，false:不能为空值
    }

    @ExceptionHandler(IllegalArgumentException.class)
    @ResponseBody
    public JsonResult handleIllegalArgumentException(Exception e, HttpServletRequest request) {
        log.error("全局异常捕捉：handleIllegalArgumentException[url:{},message:{}]", request.getRequestURI(), e.getMessage(), e);
        return JsonResult.error(e.getMessage());
    }

    @ExceptionHandler(ServletRequestBindingException.class)
    @ResponseBody
    public JsonResult handleServletRequestBindingException(ServletRequestBindingException e,
                                                           HttpServletRequest request) {
        log.error("全局异常捕捉：BindingException[url:{},message:{}]", request.getRequestURI(), e.getMessage(), e);
        return JsonResult.error(ExceptionCodeEnum.PARAMS_ILLEGAL);
    }

    @ExceptionHandler(BizException.class)
    @ResponseBody
    public JsonResult handleServiceException(BizException e, HttpServletRequest request) {
        log.error("全局异常捕捉：BizException[url:{},message:{}]", request.getRequestURI(), e.getMessage(), e);
        return JsonResult.error(e.getErrorEnum(),e.getMessage());
    }

    @ExceptionHandler(HttpRequestMethodNotSupportedException.class)
    @ResponseBody
    public JsonResult handleHttpRequestMethodNotSupportedException(HttpRequestMethodNotSupportedException e,
                                                                   HttpServletRequest request) {
        log.error("全局异常捕捉：HttpRequestMethodNotSupportedException[url:{},message:{}]", request.getRequestURI(),
                e.getMessage(), e);
        return JsonResult.error(ExceptionCodeEnum.HTTP_METHOD_ERROR);
    }

    @ExceptionHandler(MethodArgumentNotValidException.class)
    @ResponseBody
    public JsonResult handleMethodArgumentNotValidException(MethodArgumentNotValidException e, HttpServletRequest request) {
        List<FieldError> fieldErrors = e.getBindingResult().getFieldErrors();
        StringBuilder sb = new StringBuilder();
        for (FieldError fieldError : fieldErrors) {
            sb.append(fieldError.getDefaultMessage()).append(";");
        }
        log.error("全局异常捕捉：handleMethodArgumentNotValidException[url:{},message:{}]", request.getRequestURI(), e.getMessage(), e);
        return JsonResult.error(ExceptionCodeEnum.PARAMS_ILLEGAL,sb.toString());
    }

    @ExceptionHandler(MethodArgumentTypeMismatchException.class)
    @ResponseBody
    public JsonResult handleMethodArgumentTypeMismatchException(MethodArgumentTypeMismatchException e,
                                                                HttpServletRequest request) {
        log.error("全局异常捕捉：handleMethodArgumentTypeMismatchException[url:{},message:{}]", request.getRequestURI(), e.getMessage(), e);
        return JsonResult.error(ExceptionCodeEnum.PARAMS_ILLEGAL);
    }

    @ExceptionHandler(BindException.class)
    @ResponseBody
    public JsonResult handleBindException(BindException e, HttpServletRequest request) {
        log.error("全局异常捕捉：handleException[url:{},message:{}]", request.getRequestURI(), e.getMessage(), e);
        return JsonResult.error(ExceptionCodeEnum.PARAMS_ILLEGAL);
    }

    @ExceptionHandler(Exception.class)
    @ResponseBody
    public JsonResult handleException(Exception e, HttpServletRequest request) {
        log.error("全局异常捕捉：handleException[url:{},message:{}]", request.getRequestURI(), e.getMessage(), e);
        return JsonResult.error(ExceptionCodeEnum.NOT_FOUND_ERROR);
    }
}
